//
//  main.m
//  octalConversion
//
//  Created by Kevin Bradley on 6/28/18.
//  Copyright © 2018 nito. All rights reserved.
//
#import <Foundation/Foundation.h>
#include <stdio.h>
#include <errno.h>
#include <libgen.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#import <sys/utsname.h>
#import "HelperClass.h"

#define OPTION_FLAGS "o:i:ld:hcrbp:su:"

char *progname;
char *path;

static struct option longopts[] = {
    { "octal",                     required_argument,      NULL,   'o' },
    { "input",                     required_argument,      NULL,   'i' },
    { "list",                      no_argument,            NULL,   'l' },
    { "delete",                    required_argument,      NULL,   'd' },
    { "help",                      no_argument,            NULL,   'h' },
    { "clean",                     no_argument,            NULL,   'c' },
    { "repackage",                 no_argument,            NULL,   'r' },
    { "bump",                      no_argument,            NULL,   'b' },
    { "prefix",                    required_argument,      NULL,   'p' },
    { "skip",                      no_argument,            NULL,   's' },
    { "unique",                    required_argument,      NULL,   'u' },
    { NULL,                        0,                      NULL,    0  }
};

void cmd_help(void){
    printf("Usage: bootstrapTool [OPTIONS] BOOTSTRAP_FOLDER\n");
    printf("Makes various modifications to a raw bootstrap folder\n\n");
    
    printf("  -h, --help\t\t\tprints usage information\n");
    printf("  -i, --input\t\t\tthe deb package to process\n");
    printf("  -d, --delete\t\t\tthe package to delete\n");
    printf("  -l, --list\t\t\tlist all the packages installed on this bootstrap\n");
    printf("  -c, --clean\t\t\tclean a bad status file from duplicate entries\n");
    printf("  -r, --repackage\t\trepackage to fix bad locations and to change architecture\n");
    printf("  -b, --bump\t\t\trepackage and bump by one - number (ie 1.0-1 -> 1.0-2)\n");
    printf("  -s, --skip\t\t\tskip code signature checks when bumping and repackaging\n");
    printf("  -p, --prefix\t\t\tthe prefix to append to all packages when repackaging or bumping a version number\n");
    printf("\n");
}

int main(int argc, char **argv) {
    @autoreleasepool {
        
        setuid(0);
        setgid(0);
        progname = basename(argv[0]);
        path = dirname(argv[0]);
        int flag;
        NSString *octalFile = nil;
        NSString *debFile = nil;
        NSString *bootstrapPath = nil;
        NSString *deletePackage = nil;
        NSString *prefix = nil;
        BOOL listPackage = FALSE;
        BOOL repackage = FALSE;
        BOOL skip = FALSE;
        BOOL bump = FALSE;
        BOOL clean = FALSE;
        while ((flag = getopt_long(argc, argv, OPTION_FLAGS, longopts, NULL)) != -1) {
            switch(flag) {
                case 'o':
                    octalFile = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 'i':
                    debFile = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 'l':
                    listPackage = TRUE;
                    break;
                    
                case 'd':
                    deletePackage = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 'b':
                    bump = TRUE;
                    break;
                    
                case 'r':
                    repackage = TRUE;
                    break;
                    
                case 'c':
                    clean = TRUE;
                    break;
                    
                case 'p':
                    prefix = [NSString stringWithUTF8String:optarg];
                    break;
                    
                case 's':
                    skip = TRUE;
                    break;
                
                case 'u':
                    up = [NSString stringWithUTF8String:optarg];
                    break;
                    
                default:
                    cmd_help();
                    return -1;
            }
        }
        
        if (argc-optind == 1) {
            argc -= optind;
            argv += optind;
            bootstrapPath = [NSString stringWithUTF8String:argv[0]];
            DLog(@"bootstrap path: %@?", bootstrapPath);
            //return 0;
        }
        
        if (up) {
            NSArray *contents = [FM contentsOfDirectoryAtPath:up error:nil];
            NSString *newFile = [[up lastPathComponent] stringByAppendingPathExtension:@"plist"];
            NSArray *unique = [contents uniquePackageNames];
            [unique writeToFile:newFile atomically:true];
            
            DLog(@"%@", unique);
            return 0;
        }
        
        if (bump == TRUE && debFile) {
            DLog(@"\n [INFO] Bumping version number for file: %@...\n", debFile);
            InputPackage *output = [HelperClass packageForDeb:debFile];
            output.appendedPrefix = prefix;
            output.skipSignatureChecks = skip;
            [output bumpVersionInCurrentDirectory];
            return 0;
        }
        
        if (repackage == TRUE && debFile) {
            DLog(@"\n [INFO] Repackaging file: %@...\n", debFile);
            InputPackage *output = [HelperClass packageForDeb:debFile];
            output.appendedPrefix = prefix;
            output.skipSignatureChecks = skip;
            [output repackageInCurrentDirectoryWithArch:@"appletvos-arm64"];
            return 0;
        }
        
        //DLog(@"folder: %@ project: %@", folder, project);
        if (clean && bootstrapPath) {
            NSString *statusFile = [bootstrapPath relativeStatusFilePath];
            NSString *listContents = [NSString stringWithContentsOfFile:statusFile encoding:NSUTF8StringEncoding error:nil];
            //clean up any errant spaces
            listContents = [listContents stringByReplacingOccurrencesOfString:@"\n\n\n" withString:@"\n\n"];
            listContents = [listContents stringByReplacingOccurrencesOfString:@"\n\n\n\n" withString:@"\n\n"];
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
            [listContents writeToFile:statusFile atomically:TRUE];
#pragma clang diagnostic pop
            DLog(@"\n [INFO] Cleaning status file: %@", statusFile);
            
            NSString *testStatusFile = [bootstrapPath relativeStatusFilePath];
            NSArray <StatusPackageModel*> *installedPackages = [HelperClass statusInstalledPackagesFromFile:statusFile];
            __block NSMutableString *newStatusFile = [NSMutableString new];
            __block NSMutableArray *alreadyAdded = [NSMutableArray new];
            
            [installedPackages enumerateObjectsUsingBlock:^(StatusPackageModel * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
                DLog(@"\n [INFO] Processing package: %@", obj.package);
                if (![alreadyAdded containsObject:obj.package]){
                    //duplicate check first
                    NSArray *doubleCheck = [installedPackages filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"package == %@", obj.package]];
                    if (doubleCheck.count > 1){
                        NSSortDescriptor *sortDesc = [NSSortDescriptor sortDescriptorWithKey:@"version" ascending:FALSE];
                        StatusPackageModel *latestVersion = [[doubleCheck sortedArrayUsingDescriptors:@[sortDesc]] firstObject];
                        DLog(@"\n [INFO] We found a duplicate package, choosing latestVersion: %@", latestVersion);
                        [newStatusFile appendFormat:@"%@\n\n", latestVersion.rawString];
                        [alreadyAdded addObject:obj.package];
                    } else {
                        [newStatusFile appendFormat:@"%@\n\n", obj.rawString];
                    }
                } else {
                    DLog(@"\n [INFO] We already had this package, skipping older duplicate: %@\n", obj.package);
                }
            }];
            testStatusFile = [testStatusFile stringByReplacingOccurrencesOfString:@"\n\n\n" withString:@"\n\n"];
            DLog(@"\n Writing new output file: %@\n", testStatusFile);
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
            [newStatusFile writeToFile:testStatusFile atomically:TRUE];
#pragma clang diagnostic pop
            DLog(@"\n Done!\n\n");
            return 0;
            
        } else if (clean) {
            DLog(@"\nYou need to choose a bootstrap path to clean!\n\n");
            return 1;
        }
        if (deletePackage && bootstrapPath) {
            __block NSString *postRmFile = nil;
            NSString *statusFile = [bootstrapPath relativeStatusFilePath];
            NSArray <StatusPackageModel*> *installedPackages = [HelperClass statusInstalledPackagesFromFile:statusFile];
            StatusPackageModel *model = [[installedPackages filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"package == %@", deletePackage]] lastObject];
            if (model) {
                NSString *listName = [NSString stringWithFormat:@"%@.list", deletePackage];
                NSString *infoPath = [bootstrapPath stringByAppendingPathComponent:@"Library/dpkg/info"];
                NSString *listFile = [infoPath stringByAppendingPathComponent:listName];
                NSArray *listFiles = [FM contentsOfDirectoryAtPath:infoPath error:nil];
                NSPredicate *blockPred = [NSPredicate predicateWithBlock:^BOOL(id  _Nullable evaluatedObject, NSDictionary<NSString *,id> * _Nullable bindings) {
                    if ([evaluatedObject containsString:deletePackage]){
                        return TRUE;
                    }
                    return FALSE;
                }];
                NSArray *relatedFiles = [listFiles filteredArrayUsingPredicate:blockPred]; //any files in dpkg/info/deletePackage.*
                DLog(@"\nWe have found the package, determining list details from: %@", listFile);
                NSString *listContents = [NSString stringWithContentsOfFile:listFile encoding:NSUTF8StringEncoding error:nil];
                DLog(@"\nContents: %@\n", listContents);
                DLog(@"\nFound package info files: %@\n", relatedFiles);
                NSArray *files = [listContents componentsSeparatedByString:@"\n"];
                __block NSMutableArray *filesToProcess = [NSMutableArray new];
                [relatedFiles enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
                    NSString *fullPath = [infoPath stringByAppendingPathComponent:obj];
                    if([FM fileExistsAtPath:fullPath]){
                        if ([[obj pathExtension] isEqualToString:@"postrm"]){
                            postRmFile = fullPath;
                        }
                        [filesToProcess addObject:fullPath];
                    }
                }];
                [files enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
                    NSString *fullPath = [bootstrapPath stringByAppendingPathComponent:obj];
                    BOOL isDir = FALSE;
                    if([FM fileExistsAtPath:fullPath isDirectory:&isDir]){
                        if (!isDir){
                            DLog(@"found file: %@", fullPath);
                            [filesToProcess addObject:fullPath];
                        }
                    }
                }];
                DLog(@"\nFiles to process: %@\n", filesToProcess);
                NSString *error = [NSString stringWithFormat:@" [WARNING] We are about to delete '%@' from the bootstrap, this cannot be undone!", deletePackage];
                if(![HelperClass shouldContinueWithError:error]){
                    return -1;
                }
                [filesToProcess enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
                    [FM removeItemAtPath:obj error:nil];
                }];
                
                //TODO: delete the files here!
                NSString *statusFile = [bootstrapPath relativeStatusFilePath];
                [FM copyItemAtPath:statusFile toPath:[statusFile stringByAppendingPathExtension:@"bak"] error:nil];
                __block NSMutableString *newStatusFile = [NSMutableString new];
                [installedPackages enumerateObjectsUsingBlock:^(StatusPackageModel * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
                    if (obj != model){
                        [newStatusFile appendFormat:@"%@\n\n", obj.rawString];
                    } else {
                        DLog(@"skipping the model we want to delete: %@", obj);
                    }
                }];
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"
                [newStatusFile writeToFile:statusFile atomically:TRUE];
#pragma clang diagnostic pop
                if (postRmFile) {
                    DLog(@"\n [WARNING] We found a postrm file, will not run this due to potential unexpected consequences off of the target device, displaying contents so you can determine if any additional steps are necessary\n");
                    DLog(@"\n\n%@ contents:\n%@\n\n", postRmFile, [NSString stringWithContentsOfFile:postRmFile encoding:NSUTF8StringEncoding error:nil]);
                }
            }
            return 0;
        } else if (deletePackage) {
            DLog(@"\nYou need to choose a bootstrap path to delete from!\n\n");
            return 1;
        }
        if (octalFile) {
            if([FM fileExistsAtPath:octalFile]) {
                NSDictionary *attrs = [FM attributesOfItemAtPath:octalFile error:nil];
                NSNumber *posixPerms = attrs[NSFilePosixPermissions];
                DLog(@"perms: %@", posixPerms);
                return 0;
            }
            NSString *octal = [HelperClass octalFromSymbols:octalFile];
            DLog(@"%@", octal);
            return 0;
        }
        if (listPackage && bootstrapPath) {
            NSString *statusFile = [bootstrapPath relativeStatusFilePath];
            
            NSArray *installedPackages = [HelperClass statusInstalledPackagesFromFile:statusFile];
            DLog(@"%@", installedPackages);
            return 0;
            
        } else if (listPackage) {
            
            DLog(@"\nYou need to choose a bootstrap path to list packages from!\n\n");
            return 1;
        }
        
        if (debFile && bootstrapPath) {
            DLog(@"\nProcessing file: %@\n", debFile);
            InputPackage *output = [HelperClass packageForDeb:debFile];
            return [output installToBootstrapPath:bootstrapPath];
            
        } else if (debFile) {
            DLog(@"\nYou need to choose a bootstrap path to install the package to!\n\n");
            return 1;
        }
        //if we got this far.. nothing happened, print help.
        cmd_help();
        
    }
    return 0;
}
